#include "processes_page.h"
#include "ui_processes_page.h"
#include "utilities.h"
#include <unistd.h>
#include <KAuth/KAuthAction>
#include <KAuth/KAuthExecuteJob>

ProcessesPage::~ProcessesPage()
{
    delete ui;
}

ProcessesPage::ProcessesPage(QWidget *parent) :
  QWidget(parent),
  ui(new Ui::ProcessesPage),
  mItemModel(new QStandardItemModel(this)),
  mSortFilterModel(new QSortFilterProxyModel(this)),
  im(InfoManager::ins()),
  mTimer(new QTimer(this)),
  selectaction(nullptr)
{
    ui->setupUi(this);

    init();
}

void ProcessesPage::runKAuthAction(const QString& actionId, const int pids, const QVariantMap& options)
{
    KAuth::Action action(actionId);
    if (!action.isValid()) {
        return ;
    }
//    action.setParentWidget(widget);
    action.setHelperId(QStringLiteral("org.nde.auth.helper"));

//    const int processCount = pids.count();
//    for (int i = 0; i < processCount; ++i) {
//        action.addArgument(QStringLiteral("pid").arg(i), pids.at(i));
//    }
    action.addArgument(QStringLiteral("pid"), pids);
//    action.addArgument(QStringLiteral("pidcount"), processCount);

    for (auto itr = options.cbegin(); itr != options.cend(); ++itr) {
        action.addArgument(itr.key(), itr.value());
    }

    KAuth::ExecuteJob *job = action.execute();
    if(job->exec()) {
        return ;
    } else {
        return ;
    }
}

bool ProcessesPage::eventFilter(QObject* obj,QEvent *evt)
{
    if(obj == ui->tableProcess && evt->type() ==QEvent::ContextMenu){
        if(ui->tableProcess->currentIndex().isValid()){
            QModelIndexList selecteds = ui->tableProcess->selectionModel()->selectedRows();
            if (! selecteds.isEmpty()) {
                mSeletedRowModel = selecteds.first();
            } else {
                mSeletedRowModel = QModelIndex();
            }

            pid_t pid = mSeletedRowModel.data(1).toInt();
            initPriority(pid);
            mMenu.exec(cursor().pos());
        }
    }
    return QWidget::eventFilter(obj,evt);
}

void ProcessesPage::init()
{
    mHeaders = QStringList {
        "PID", tr("Resident Memory"), tr("%Memory"), tr("Virtual Memory"),
        tr("User"), "%CPU", tr("Start Time"), tr("State"), tr("Group"),
        tr("Nice"), tr("CPU Time"), tr("Session"), tr("Process")
    };

    // slider settings
    ui->sliderRefresh->setRange(1, 10);
    ui->sliderRefresh->setPageStep(1);
    ui->sliderRefresh->setSingleStep(1);

    // Table settings
    mSortFilterModel->setSourceModel(mItemModel);

    mItemModel->setHorizontalHeaderLabels(mHeaders);

    ui->tableProcess->setModel(mSortFilterModel);
    mSortFilterModel->setSortRole(1);
    mSortFilterModel->setDynamicSortFilter(true);
    mSortFilterModel->sort(5, Qt::SortOrder::DescendingOrder);

    ui->tableProcess->horizontalHeader()->setSectionsMovable(true);
    ui->tableProcess->horizontalHeader()->setFixedHeight(36);
    ui->tableProcess->horizontalHeader()->setDefaultAlignment(Qt::AlignLeft | Qt::AlignVCenter);
    ui->tableProcess->horizontalHeader()->setCursor(Qt::PointingHandCursor);
    ui->tableProcess->horizontalHeader()->resizeSection(0, 70);



    loadMenu();
    loadProcesses();

    connect(mTimer, &QTimer::timeout, this, &ProcessesPage::loadProcesses);
    mTimer->setInterval(1000);
    mTimer->start();

    ui->tableProcess->horizontalHeader()->setContextMenuPolicy(Qt::CustomContextMenu);

    connect(ui->tableProcess->horizontalHeader(), SIGNAL(customContextMenuRequested(const QPoint&)),
        this, SLOT(on_tableProcess_customContextMenuRequested(const QPoint&)));

    loadHeaderMenu();

    Utilities::addDropShadow(ui->btnEndProcess, 60);
    Utilities::addDropShadow(ui->tableProcess, 55);
    ui->tableProcess->installEventFilter(this);
}
void ProcessesPage::initPriority(pid_t _pid)
{
    if(selectaction){
        selectaction->setChecked(false);
    }
    int scheduler = sched_getscheduler(_pid);
    struct sched_param param;
    sched_getparam(_pid,&param);
    int maxpri = sched_get_priority_max(scheduler);
    switch (scheduler) {
    case SCHED_FIFO:{
        if(param.sched_priority <= maxpri / 5){
            selectaction = fifo_act_low;
        } else if(param.sched_priority <= maxpri * 2/ 5){
            selectaction = fifo_act_tanlow;
        } else if(param.sched_priority <= maxpri * 3/ 5){
            selectaction = fifo_act_normal;
        } else if(param.sched_priority <= maxpri * 4/ 5){
            selectaction = fifo_act_lesshight;
        } else if(param.sched_priority <= maxpri){
            selectaction = fifo_act_hight;
        }
        break;
        }
    case SCHED_RR:{
        if(param.sched_priority <= maxpri / 5){
            selectaction = rr_act_low;
        } else if(param.sched_priority <= maxpri * 2/ 5){
            selectaction = rr_act_tanlow;
        } else if(param.sched_priority <= maxpri * 3/ 5){
            selectaction = rr_act_normal;
        } else if(param.sched_priority <= maxpri * 4/ 5){
            selectaction = rr_act_lesshight;
        } else if(param.sched_priority <= maxpri){
            selectaction = rr_act_hight;
        }
        break;
        }
    default:
        selectaction = normal_action;
        break;
    }
    selectaction->setChecked(true);
}
void ProcessesPage::loadMenu()
{

    priomenu = new QMenu(this);
    priomenu->setTitle(tr("set scheduler"));

    normal_action = new QAction(tr("Normal"));
    normal_action->setCheckable(true);
    connect(normal_action,SIGNAL(triggered()),this,SLOT(on_ProcessnormalPriority()));
    priomenu->addAction(normal_action);

    menufifo = new QMenu(this);
    menufifo->setTitle(tr("FIFO"));

    fifo_act_low = new QAction(tr("low"));
    fifo_act_low->setCheckable(true);
    connect(fifo_act_low,SIGNAL(triggered()),this,SLOT(on_ProcessfifoPrioritylow()));
    menufifo->addAction(fifo_act_low);

    fifo_act_tanlow = new QAction(tr("thanlow"));
    fifo_act_tanlow->setCheckable(true);
    connect(fifo_act_tanlow,SIGNAL(triggered()),this,SLOT(on_ProcessfifoPrioritylanlow()));
    menufifo->addAction(fifo_act_tanlow);

    fifo_act_normal = new QAction(tr("normal"));
    fifo_act_normal->setCheckable(true);
    connect(fifo_act_normal,SIGNAL(triggered()),this,SLOT(on_ProcessfifoPrioritynormal()));
    menufifo->addAction(fifo_act_normal);

    fifo_act_lesshight = new QAction(tr("LessHight"));
    fifo_act_lesshight->setCheckable(true);
    connect(fifo_act_lesshight,SIGNAL(triggered()),this,SLOT(on_ProcessfifoPrioritylesshight()));
    menufifo->addAction(fifo_act_lesshight);

    fifo_act_hight = new QAction(tr("Hight"));
    fifo_act_hight->setCheckable(true);
    connect(fifo_act_hight,SIGNAL(triggered()),this,SLOT(on_ProcessfifoPriorityhight()));
    menufifo->addAction(fifo_act_hight);

    priomenu->addMenu(menufifo);

    menuRR = new QMenu(this);
    menuRR->setTitle(tr("RR"));

    rr_act_low = new QAction(tr("low"));
    rr_act_low->setCheckable(true);
    connect(rr_act_low,SIGNAL(triggered()),this,SLOT(on_ProcessrrPrioritylow()));
    menuRR->addAction(rr_act_low);

    rr_act_tanlow = new QAction(tr("thanlow"));
    rr_act_tanlow->setCheckable(true);
    connect(rr_act_tanlow,SIGNAL(triggered()),this,SLOT(on_ProcessrrPrioritylanlow()));
    menuRR->addAction(rr_act_tanlow);

    rr_act_normal = new QAction(tr("normal"));
    rr_act_normal->setCheckable(true);
    connect(rr_act_normal,SIGNAL(triggered()),this,SLOT(on_ProcessrrPrioritynormal()));
    menuRR->addAction(rr_act_normal);

    rr_act_lesshight = new QAction(tr("LessHight"));
    rr_act_lesshight->setCheckable(true);
    connect(rr_act_lesshight,SIGNAL(triggered()),this,SLOT(on_ProcessrrPrioritylesshight()));
    menuRR->addAction(rr_act_lesshight);

    rr_act_hight = new QAction(tr("Hight"));
    rr_act_hight->setCheckable(true);
    connect(rr_act_hight,SIGNAL(triggered()),this,SLOT(on_ProcessrrPriorityhight()));
    menuRR->addAction(rr_act_hight);

    priomenu->addMenu(menuRR);


    mMenu.addMenu(priomenu);
}
void ProcessesPage::loadHeaderMenu()
{
    int i = 0;
    for (const QString &header : mHeaders) {
        QAction *action = new QAction(header);
        action->setCheckable(true);
        action->setChecked(true);
        action->setData(i++);

        mHeaderMenu.addAction(action);
    }

    // exclude headers
    QList<int> hiddenHeaders = { 3, 6, 7, 8, 9, 10, 11 };

    QList<QAction*> actions = mHeaderMenu.actions();
    for (const int i : hiddenHeaders) {
        if (i < mHeaders.count()) {
            ui->tableProcess->horizontalHeader()->setSectionHidden(i, true);
            actions.at(i)->setChecked(false);
        }
    }
}

void ProcessesPage::loadProcesses()
{
    QModelIndexList selecteds = ui->tableProcess->selectionModel()->selectedRows();

    mItemModel->removeRows(0, mItemModel->rowCount());

    im->updateProcesses();

    QList<Process> processes = im->getProcesses();
    QString username = im->getUserName();

    if (ui->checkAllProcesses->isChecked()) {
        for (const Process &proc : processes) {
            mItemModel->appendRow(createRow(proc));
        }
    } else  {
        for (const Process &proc : processes) {
            if (username == proc.getUname()) {
                mItemModel->appendRow(createRow(proc));
            }
        }
    }

    ui->lblProcessTitle->setText(tr("Processes (%1)").arg(mItemModel->rowCount()));

    // selected item
    if (! selecteds.isEmpty()) {
        mSeletedRowModel = selecteds.first();

        for (int i = 0; i < mSortFilterModel->rowCount(); ++i) {
            if (mSortFilterModel->index(i, 0).data(1).toInt() == mSeletedRowModel.data(1).toInt()) {
                ui->tableProcess->selectRow(i);
            }
        }
    } else {
        mSeletedRowModel = QModelIndex();
    }
}

QList<QStandardItem*> ProcessesPage::createRow(const Process &proc)
{
    QList<QStandardItem*> row;

    int data = 1;

    QStandardItem *pid_i = new QStandardItem(QString::number(proc.getPid()));
    pid_i->setData(proc.getPid(), data);
    pid_i->setData(proc.getPid(), Qt::ToolTipRole);

    QStandardItem *rss_i = new QStandardItem(FormatUtil::formatBytes(proc.getRss()));
    rss_i->setData(proc.getRss(), data);
    rss_i->setData(FormatUtil::formatBytes(proc.getRss()), Qt::ToolTipRole);

    QStandardItem *pmem_i = new QStandardItem(QString::number(proc.getPmem()));
    pmem_i->setData(proc.getPmem(), data);
    pmem_i->setData(proc.getPmem(), Qt::ToolTipRole);

    QStandardItem *vsize_i = new QStandardItem(FormatUtil::formatBytes(proc.getVsize()));
    vsize_i->setData(proc.getVsize(), data);
    vsize_i->setData(FormatUtil::formatBytes(proc.getVsize()), Qt::ToolTipRole);

    QStandardItem *uname_i = new QStandardItem(proc.getUname());
    uname_i->setData(proc.getUname(), data);
    uname_i->setData(proc.getUname(), Qt::ToolTipRole);

    QStandardItem *pcpu_i = new QStandardItem(QString::number(proc.getPcpu()));
    pcpu_i->setData(proc.getPcpu(), data);
    pcpu_i->setData(proc.getPcpu(), Qt::ToolTipRole);

    QStandardItem *starttime_i = new QStandardItem(proc.getStartTime());
    starttime_i->setData(proc.getStartTime(), data);
    starttime_i->setData(proc.getStartTime(), Qt::ToolTipRole);

    QStandardItem *state_i = new QStandardItem(proc.getState());
    state_i->setData(proc.getState(), data);
    state_i->setData(proc.getState(), Qt::ToolTipRole);

    QStandardItem *group_i = new QStandardItem(proc.getGroup());
    group_i->setData(proc.getGroup(), data);
    group_i->setData(proc.getGroup(), Qt::ToolTipRole);

    QStandardItem *nice_i = new QStandardItem(QString::number(proc.getNice()));
    nice_i->setData(proc.getNice(), data);
    nice_i->setData(proc.getNice(), Qt::ToolTipRole);

    QStandardItem *cpuTime_i = new QStandardItem(proc.getCpuTime());
    cpuTime_i->setData(proc.getCpuTime(), data);
    cpuTime_i->setData(proc.getCpuTime(), Qt::ToolTipRole);

    QStandardItem *session_i = new QStandardItem(proc.getSession());
    session_i->setData(proc.getSession(), data);
    session_i->setData(proc.getSession(), Qt::ToolTipRole);

    QStandardItem *cmd_i = new QStandardItem(proc.getCmd());
    cmd_i->setData(proc.getCmd(), data);
    cmd_i->setData(QString("<p>%1</p>").arg(proc.getCmd()), Qt::ToolTipRole);

    row << pid_i << rss_i << pmem_i << vsize_i << uname_i << pcpu_i
        << starttime_i << state_i << group_i << nice_i << cpuTime_i
        << session_i << cmd_i;

    return row;
}

void ProcessesPage::on_txtProcessSearch_textChanged(const QString &val)
{
    QRegExp query(val, Qt::CaseInsensitive, QRegExp::Wildcard);

    mSortFilterModel->setFilterKeyColumn(mHeaders.count() - 1); // process name
    mSortFilterModel->setFilterRegExp(query);
}

void ProcessesPage::on_sliderRefresh_valueChanged(const int &i)
{
    ui->lblRefresh->setText(tr("Refresh (%1)").arg(i));
    mTimer->setInterval(i * 1000);
}

void ProcessesPage::on_btnEndProcess_clicked()
{
    QModelIndexList selecteds = ui->tableProcess->selectionModel()->selectedRows();
    if (! selecteds.isEmpty()) {
        mSeletedRowModel = selecteds.first();
    } else {
        mSeletedRowModel = QModelIndex();
    }

    pid_t pid = mSeletedRowModel.data(1).toInt();

    if (pid) {
        QString selectedUname = mSortFilterModel->index(mSeletedRowModel.row(), 4).data(1).toString();

        try {
            if (selectedUname == im->getUserName()) {
                CommandUtil::exec("kill", { QString::number(pid) });
            } else {
                CommandUtil::sudoExec("kill", { QString::number(pid) });
            }
        } catch (QString &ex) {
            qCritical() << ex;
        }
    }
}

bool ProcessesPage::setscheduler(pid_t _pid,int priority,int level)
{
    runKAuthAction(QStringLiteral("org.nde.auth.helper.changecpuscheduler"),
                    _pid,
                    {{QStringLiteral("cpuScheduler"), priority}, {QStringLiteral("cpuSchedulerPriority"), level}}
                );
    return false;
}



void ProcessesPage::on_ProcessnormalPriority()
{
    if(selectaction){
        selectaction->setChecked(false);
    }
    selectaction = normal_action;
    pid_t pid = mSeletedRowModel.data(1).toInt();
    setscheduler(pid,SCHED_OTHER,0);
}

void ProcessesPage::on_ProcessfifoPrioritylow()
{
    if(selectaction){
        selectaction->setChecked(false);
    }
    selectaction = fifo_act_low;
    pid_t pid = mSeletedRowModel.data(1).toInt();
    int maxpri = sched_get_priority_max(SCHED_FIFO);
    setscheduler(pid,SCHED_FIFO,maxpri/5);
}

void ProcessesPage::on_ProcessfifoPrioritylanlow()
{
    if(selectaction){
        selectaction->setChecked(false);
    }
    selectaction = fifo_act_tanlow;
    pid_t pid = mSeletedRowModel.data(1).toInt();
    int maxpri = sched_get_priority_max(SCHED_FIFO);
    setscheduler(pid,SCHED_FIFO,maxpri*2/5);
}
void ProcessesPage::on_ProcessfifoPrioritynormal()
{
    if(selectaction){
        selectaction->setChecked(false);
    }
    selectaction = fifo_act_normal;
    pid_t pid = mSeletedRowModel.data(1).toInt();
    int maxpri = sched_get_priority_max(SCHED_FIFO);
    setscheduler(pid,SCHED_FIFO,maxpri*3/5);
}
void ProcessesPage::on_ProcessfifoPrioritylesshight()
{
    if(selectaction){
        selectaction->setChecked(false);
    }
    selectaction = fifo_act_lesshight;
    pid_t pid = mSeletedRowModel.data(1).toInt();
    int maxpri = sched_get_priority_max(SCHED_FIFO);
    setscheduler(pid,SCHED_FIFO,maxpri*4/5);
}
void ProcessesPage::on_ProcessfifoPriorityhight()
{
    if(selectaction){
        selectaction->setChecked(false);
    }
    selectaction = fifo_act_hight;
    pid_t pid = mSeletedRowModel.data(1).toInt();
    int maxpri = sched_get_priority_max(SCHED_FIFO);
    setscheduler(pid,SCHED_FIFO,maxpri);
}
void ProcessesPage::on_ProcessrrPrioritylow()
{
    if(selectaction){
        selectaction->setChecked(false);
    }
    selectaction = rr_act_low;
    pid_t pid = mSeletedRowModel.data(1).toInt();
    int maxpri = sched_get_priority_max(SCHED_RR);
    setscheduler(pid,SCHED_RR,maxpri/5);
}
void ProcessesPage::on_ProcessrrPrioritylanlow()
{
    if(selectaction){
        selectaction->setChecked(false);
    }
    selectaction = rr_act_tanlow;
    pid_t pid = mSeletedRowModel.data(1).toInt();
    int maxpri = sched_get_priority_max(SCHED_RR);
    setscheduler(pid,SCHED_RR,maxpri * 2/5);
}
void ProcessesPage::on_ProcessrrPrioritynormal()
{
    if(selectaction){
        selectaction->setChecked(false);
    }
    selectaction = rr_act_normal;
    pid_t pid = mSeletedRowModel.data(1).toInt();
    int maxpri = sched_get_priority_max(SCHED_RR);
    setscheduler(pid,SCHED_RR,maxpri * 3/5);
}
void ProcessesPage::on_ProcessrrPrioritylesshight()
{
    if(selectaction){
        selectaction->setChecked(false);
    }
    selectaction = rr_act_lesshight;
    pid_t pid = mSeletedRowModel.data(1).toInt();
    int maxpri = sched_get_priority_max(SCHED_RR);
    setscheduler(pid,SCHED_RR,maxpri * 4/5);
}
void ProcessesPage::on_ProcessrrPriorityhight()
{
    if(selectaction){
        selectaction->setChecked(false);
    }
    selectaction = rr_act_hight;
    pid_t pid = mSeletedRowModel.data(1).toInt();
    int maxpri = sched_get_priority_max(SCHED_RR);
    setscheduler(pid,SCHED_RR,maxpri);
}

void ProcessesPage::on_tableProcess_customContextMenuRequested(const QPoint &pos)
{
    QPoint globalPos = ui->tableProcess->mapToGlobal(pos);
    QAction *action = mHeaderMenu.exec(globalPos);
    if (action) {
        ui->tableProcess->horizontalHeader()->setSectionHidden(action->data().toInt(), ! action->isChecked());
    }


}
